/* SPDX-License-Identifier: GPL-2.0-only OR License-Ref-kk-custom */
/*
 * Copyright (C) 2022 Kernkonzept GmbH.
 * Author(s): Stephan Gerhold <stephan.gerhold@kernkonzept.com>
 */

#include <cstdlib>
#include <cstring>
#include <algorithm>

#include <lwip/dhcp.h>
#include <lwip/ip4_addr.h>
#include <lwip/netif.h>
#include <lwip/tcpip.h>

enum
{
  Max_env_len = sizeof("IFCONFIG_IP4_") + NETIF_NAMESIZE,
};

struct Ifconfig_env
{
  Ifconfig_env();
  static void run(void*);
  static void configure(netif *netif, const char *ifname);
  static void configure_ip4(netif *netif, const char *val);
};

Ifconfig_env::Ifconfig_env()
{
  tcpip_callback_wait(run, nullptr);
}

void Ifconfig_env::run(void*)
{
  char name[NETIF_NAMESIZE];
  netif *netif;

  NETIF_FOREACH(netif)
    configure(netif, netif_index_to_name(netif_get_index(netif), name));
}

void Ifconfig_env::configure(netif *netif, const char *ifname)
{
  char env[Max_env_len] = "IFCONFIG_IP4_";
  strcat(env, ifname);

  auto val = getenv(env);
  if (!val)
    return;

  configure_ip4(netif, val);
}

static u8_t *ip4_addr_p(ip4_addr_t *ip, int idx)
{ return &reinterpret_cast<u8_t *>(&ip->addr)[idx]; }

void Ifconfig_env::configure_ip4(netif *netif, const char *val)
{
  // IFCONFIG_IP4_vn0 "192.168.42.2/24 via 192.168.42.1" or "dhcp"
  if (strcmp(val, "dhcp") == 0)
    {
      dhcp_start(netif);
      return;
    }

  ip4_addr_t ip, netmask, gw;
  auto prefix = 32U;

  auto n = sscanf(val, "%hhu.%hhu.%hhu.%hhu/%u via %hhu.%hhu.%hhu.%hhu",
                  ip4_addr_p(&ip, 0), ip4_addr_p(&ip, 1),
                  ip4_addr_p(&ip, 2), ip4_addr_p(&ip, 3),
                  &prefix,
                  ip4_addr_p(&gw, 0), ip4_addr_p(&gw, 1),
                  ip4_addr_p(&gw, 2), ip4_addr_p(&gw, 3));

  if (n <= 8)
    ip4_addr_set_any(&gw);
  if (n < 4)
    ip4_addr_set_any(&ip);
  ip4_addr_set_u32(&netmask, lwip_htonl(~0UL << (32U - std::min(prefix, 32U))));

  netif_set_addr(netif, &ip, &netmask, &gw);
}

Ifconfig_env ifce;
